#!/usr/bin/perl
# Backup Script for FromTo.Cc
# (C) 2009 by Tatsumi Hosokawa <hosokawa@FromTo.Cc>

use strict;

my $rm = '/bin/rm';
my $mount = '/sbin/mount';
my $umount = '/sbin/umount';
my $mksnap_ffs = '/sbin/mksnap_ffs';
my $mdconfig = '/sbin/mdconfig';
my $mdnum = '8';
my $rsync = '/usr/local/bin/rsync';
my $find = '/usr/bin/find';

my $rootfs = '/backup/root';
my $snapfsbackbase = '/backup';
my $snapfssrcbase = '/backup2';
my @snapfslist = ('/tmp', '/var', '/usr', '/data');
my @excluderootlist = ('.snap', 'backup', 'dev', 'tmp', 'var', 'usr', 'data');

print STDERR "mounting filesystems\n";

if (psystem("$mount $rootfs")) {
    umountall();
    die "cannot mount $rootfs";
}
foreach my $i (@snapfslist) {
    my $backdir = "$snapfsbackbase$i";
    if (psystem("$mount $backdir")) {
	umountall();
	die "cannot mount $rootfs";
    }
}

print STDERR "mounted!\n";
print STDERR "backup root filesystem\n";

my $exclude = "/tmp/backup-pl.$$";

open my $excludefile, "> $exclude" or die "Cannot open exlucde file";
print $excludefile <<EOT;
- /dev
- /.snap
EOT

foreach my $i (@snapfslist) {
    print $excludefile "- $i/\n";
}

close $excludefile;

my @rootfiles = ((glob "/*"), (glob "/.??*"));
my $rootfiles = '';

ROOT: foreach my $i (@rootfiles) {
    $i =~ s!^/!!;
    foreach my $j (@excluderootlist) {
	next ROOT if $i eq $j;
    }
    $rootfiles .= "$i ";
}

psystem("cd / ; $rsync -avH --delete --numeric-ids --filter='. $exclude' $rootfiles $rootfs");

foreach my $i (@snapfslist) {
    my $snapimage = "$i/.snap/snapshot";
    my $snapfssrc = "$snapfssrcbase$i";
    if (-f $snapimage) {
	umountall();
	die "$snapimage exists";
    }
    print STDERR "Creating snapshot for $i\n";
    psystem("$mksnap_ffs $i $snapimage");
    psystem("$mdconfig -a -t vnode -o readonly -f $snapimage -u $mdnum");
    psystem("$mount -o ro /dev/md$mdnum $snapfssrc");
    print STDERR "Mounted snapshot for $i on $snapfssrc\n";
    my $backdir = "$snapfsbackbase$i";
    open my $excludefile, "> $exclude" or die "Cannot open exlucde file";
    print $excludefile <<EOT;
- $i/.snap
EOT
    open my $find, "cd $snapfssrcbase; $find $i -type b -or -type c -or -type p -or -type s |" or die "cannot open find(1)";
    while (<$find>) {
	chomp;
	print $excludefile "- /$_\n";
    }
    close $find;
    close $excludefile;
    print STDERR "Syncing $snapfssrc to $backdir\n";
    psystem("cd / ; $rsync -avH --delete --numeric-ids --filter='. $exclude' $snapfssrc $snapfsbackbase");
    print STDERR "Removing snapshot for $i\n";
    psystem("$umount $snapfssrc");
    psystem("$mdconfig -d -u $mdnum");
    psystem("rm -f $snapimage");
}

umountall();

unlink $exclude;

sub umountall () {
    foreach my $i (@snapfslist) {
	my $backdir = "$snapfsbackbase$i";
	psystem("$umount $backdir");
    }
    psystem("$umount $rootfs");
}

sub psystem ($) {
    my $cmd = shift;
    print STDERR "$cmd\n";
    return system $cmd;
}
